// The MIT License (MIT)
// 
// Copyright (c) 2014 Rafat Rashid
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

/** @file MFraction.h
	@brief This file declares the interface of the %MFraction class.

   Private and protected members of the class are also declared in this file.		   
*/

#ifndef MFRACTION_H
	#define MFRACTION_H

	#include "PartialLib.h"

	#ifndef FRACTION_ONLY		

		#include "Fraction.h"

		/**
			@brief This class allows you to handle improper as well as mixed fractions just like how basic types
			such as integers and doubles are handled in c++.

			Improper fractions are of the form numerator/denominator whereas mixed fractions of the form
			whole/numerator/denominator where numerator in this case is always less than the denominator.
			For instance, 5/6 or 4/2/3. These fractions will always be in their reduced form.

			The programmer has a choice of outputting the fraction as a decimal number, as an improper fraction
			or as a mixed fraction. This applies to all fraction objects that are of type %MFraction.
			For instance, 3.5 or 7/2 or 3/1/2. See documentation of "Static Public Member Functions" for details.

			The programmer can initilize fraction objects by providing (note the integer or string values provided
			do not have to be in their reduced fractional form):
				- a single number (can be integer or decimal value)
				- the numerator and denominator of the fraction to represent (integer values)
				- the whole, numerator and denominator of the fraction to represent (integer values)
				- a valid string value (such as "8/6", "3.5", "4" or "3/9/3")
				- another %Fraction or %MFraction object
				- see the constructor documentation below for details

			Arithmetic and comparison operations can also be performed on %MFraction objects, analogous to how
			these operations work on values of basic types such as an int.
					
			I/O with %MFractions works similarly to how it works with integers as well. For instance,
			6/4/2, 2/3, 8/4 and 5.6 are all valid inputs. These inputs will be stored in their reduced form.
		*/

		class MFraction : public Fraction
		{
			public:

				/*! @brief Enumeration type to track how objects of type %MFraction should be outputted. */
				/*! This is independent of the format properties of a fraction of type %Fraction.
					@sa MFraction::FORMAT */
				enum FracFormat
				{
					MIX_FRAC = 0, ///< = 0: output as mixed fraction when it applies, ie 5/6 and 1/2/3
					IM_FRAC = 1,  ///< = 1: output as improper fractions ie. 5/6, 8/5
					DECI = 2      ///< = 2: output as a decimal number ie. 3.5
				};

			private:
				//private allows only %MFraction class members access (and its friends)

				//! Stores how objects of type %MFraction is to be outputted.
				/*!	This applies to all objects of type %MFraction. The default behaviour is as a mixed fraction.
					Use MFraction::setFormat() to set %MFraction format and MFraction::getFormat() to get the
					format. You can set/get the following values:						

					@note Using %MFraction:: instead of %Fraction:: means you are working with the %MFraction version
						  of the FracFormat enumerated type. Similarly, using %Fraction:: means you will be setting or
						  gettting %Fraction objects' formatting, not MFraction's.			    
				*/
				static FracFormat FORMAT;

			public:

				//static functions to set and get the %MFraction format
				//static functions cant access non-static variables		

				/*! @brief Sets the format of how objects of type %MFraction is outputted.
					@param format What the format of the fraction should be:
							 		  - either MFraction::MIX_FRAC
									  - MFraction::IM_FRAC
									  - or MFraction::DECI
				*/
				static void setFormat(const FracFormat &format);
				
				/*! @brief Use to see what the current output format of fractions of type %MFraction is.
					@return Returns current output format for fractions of type %MFraction.
							This is a value of type MFraction::FracFormat.
				*/
				static FracFormat getFormat();

			private:
		
				/* THE DATA MEMBERS: */			
				
				/*! @brief Stores the whole number portion of the fraction */
				unsigned int whole;

				/* HELPER METHODS PRIVATE TO CLASS */

				/*! @brief Reduces a reduced improper fraction into a mixed fraction. */
				/*! For instance, 6/5 -> 1/1/5
					@sa Fraction::reduce() */
				void reduce();

			public:

				/*! @name Constructors/Destructor
					A constructor initilizes (creates) an object of type %MFraction.
					A destructor destroys an object of type %MFraction when it goes out of scope. */
				//@{

				//! Default Constructor: sets fraction to 0.			
				/*! This is the constructor that is used when no constructor is explicitely specifed.
					@return %MFraction object.
				*/
				MFraction();

				/*! @brief Converts a decimal number (or integer) into a fraction.
					@param number Decimal number (or integer) to convert to fraction.
					@return %MFraction object.
				*/
				MFraction(const double &number);

				/*! @brief Takes numerator and denominator of fraction (not reduced) and creates a reduced mixed fraction.
					@param numerator Numerator of fraction.
					@param denominator Denominator of fraction.
					@return %MFraction object.
					@exception If denominator is zero, throws FR_ERROR with value FR_DENOM_ZERO. Object is left
							   uncreated (also means destructor will not run as well).
				*/
				MFraction(const int &numerator, const int &denominator);

				/*! @brief Takes whole, numerator and denominator of fraction (not reduced) and creates a reduced mixed fraction.
					@param whole Whole number portion of the fraction.
					@param numerator Numerator of fraction.
					@param denominator Denominator of fraction.
					@return %MFraction object.
					@exception If denominator is zero, throws FR_ERROR with value FR_DENOM_ZERO. Object is left
							   uncreated (also means destructor will not run as well).
				*/
				MFraction(const int &whole, const int &numerator, const int &denominator);

				/*! @brief Converts a valid character array into a fraction.
					@param frac Character array to convert. Possible values include "3.4", "4/7/3" and "16/8".
				 				Invalid values include "3.4.5" and "3.4/5.6".
					@return %MFraction object.
					@exception If input is invalid, throws FR_ERROR with value FR_STR_INVALID. Object is left
							   uncreated (also means destructor will not run as well).
				*/
				MFraction(const char *frac);

				/*! @brief Copy Constructor: copies another fraction (of type %Fraction) into the created fraction
						   (of type %MFraction).
					@param frac %Fraction object to copy.
					@return %MFraction object.				
				*/
				MFraction(const Fraction &frac);

				/*! @brief Copy Constructor: copies another fraction (of type %MFraction) into the created fraction
						   (of type %MFraction).
					@param frac %MFraction object to copy.
					@return %MFraction object.				
				*/
				MFraction(const MFraction &frac);

				//! Destructor: used to destory objects of type %MFraction.
				/*! This cannot be called explicitely. */
				~MFraction();

				//@}

				/*! @name Mutator Methods
					Use these methods to set the whole/numerator/denominator independently.

					(to get the values of the whole, numerator and denominator, see MFraction::operator [])
				*/
				//@{

				//! Sets the whole number portion of the fraction.
				/*! @param whole The number to set the whole of the fraction to. */
				void setWhole(int whole);

				//! Sets the numerator of the fraction.
				/*! @param numerator The number to set the numerator of the fraction to. */
				void setNum(int numerator);

				//! Sets the denomerator of the fraction.
				/*! @param denominator The number to set the denomerator of the fraction to.
					@exception Throws FR_ERROR with value FR_DENOM_ZERO if parameter's value is 0.
				*/
				void setDen(int denominator);

				//@}

				/*! @name Member Overloaded Operators
					The following are overloaded operators which are members of the class.
					
					One of the "parameters" to the operation is implicit and is the "calling object".
					The calling object must be of type %MFraction.
					For instance, x[0] and --x, where x is the calling object.
				*/
				//@{		

				/*! @brief Subscript operator: Pass in 0 to get whole (signed), 1 to get numerator (signed),
					2 for denominator (unsigned).
				    
					This operator must be a member function.

					@param subscript value of 0, 1 or 2. Any other value will throw an exception.
					@return Returns an integer value (not a reference)
								- whole (signed) if subscript = 0
								- numerator (signed) if subscript = 1
								- denominator (unsigned) if subscript = 2
					@exception throws FR_ERROR with value FR_INDEX_OUT_BOUNDS if subscript is greater than 2.
				*/
				int operator [] (const unsigned int &subscript) const;

				//! Assignment operator: Calling object is made equal to object on the right of operator.
				/*! If a %Fraction object is on the right of the operator, it will be automatically converted
					to a %MFraction (using the available constructor) and then assigned to the calling object.
					This wont work if a MFraction object is being assigned to a Fraction object (the whole portion
					will be removed in that case). */
				
				/*  Additionally, if values of integer, double or strings are provided, they will be automatically
					converted to %MFraction objects prior to assignment.
					
					@note This operator must be declared as a member function.
					@param right %Fraction object on the right of the equal sign. Automatic type conversions
								 are made when values of integer, double or string is provided.			    
					@return Returns reference to object to the left of the equal sign. This makes it possible to
							do x = y = z.
				*/
				MFraction &operator = (const MFraction& right);

				/*! @brief Negation operator: returns the negated value of the calling %MFraction object.
					@note Doesn't change the value of the calling object. The calling object must be of type %MFraction.
					@return Returns the negated value of the calling %MFraction object.
							Note that operation doesn't return a reference.
							ie. -b=c, b wont be assigned to c.
				*/
				MFraction operator - () const;

				/*! @brief Prefix increment operator: adds 1 to fraction.
					@return Returns the updated fraction object so that c = ++b is possible, (here the numerator
							of b is incremented by 1 and then b is assigned to c).
					@note If returned reference instead, in cases such as ++d = b, d would be incremented and
						  then assigned value of b.
				*/
				MFraction operator ++ ();
				
				/*! @brief Prefix decrement operator: subtracts 1 from fraction.
					@return Returns the updated fraction object so that c = --b is possible, (here the numerator
							of b is decremented by 1 and then b is assigned to c).
					@note If returned reference instead, in cases such as --d = b, d would be decremented and
						  then assigned value of b.
				*/
				MFraction operator -- ();
				
				/*! @brief Postfix increment operator: increments fraction by inc; if inc is 0, increments by 1.
					@param inc
							 - Implicit - c++ : increments numerator by 1 (inc will be equal to 0)
							 - Explicit - MFraction::operator ++ (value) : increments numerator by value (inc = value)
					@return Returns the updated fraction object so that c = b++ is possible, (here, b is assigned to
							c and then b is incremented by 1).
					@note If returned reference instead, in cases such as d++ = b, b's value would be assigned to d
						  and then d's numerator would be incremented by 1.
				*/
				MFraction operator ++ (int inc);

				/*! @brief Postfix decrement operator: decrements fraction by dec; if dec is 0, decrements by 1.
					@param dec
							 - Implicit - c-- : decrements numerator by 1 (dec will be equal to 0)
							 - Explicit - MFraction::operator -- (value) : decrements numerator by value (dec = value)
					@return Returns the updated fraction object so that c = b-- is possible, (here, b is assigned to
							c and then b is decremented by 1).
					@note If returned reference instead, in cases such as d-- = b, b's value would be assigned to d
						  and then d's numerator would be decremented by 1.
				*/
				MFraction operator -- (int dec);

				//! Addition/assignment operator: adds fraction on the right of operator to fraction on the left.
				/*! Calling object must be of type %Mraction. The calling object is the fraction that is on the
					left of the operator.

					@param right %MFraction object on the right of operator that is to be added. If values of
						   integer, double or strings are provided, they will be automatically converted to
						   %MFraction objects prior to addition/assignment.
					@return Returns a %MFraction object that is equal to the value of the updated calling object.
				*/
				MFraction operator += (const MFraction &right);
				
				/*! @brief Subtraction/assignment operator: subtracts fraction on the right of operator from
						   fraction on the left.

					Calling object must be of type %MFraction. The calling object is the fraction that is on the
					left of the operator.
					
					@param right %MFraction object on the right of operator that is to be subtracted. If values of
						   integer, double, strings or %Fraction are provided, they will be automatically converted to
						   %MFraction objects prior to subtraction/assignment.
					@return Returns a %MFraction object that is equal to the value of the updated calling object.
				*/
				MFraction operator -= (const MFraction &right);

				/*! @brief Multiplication/assignment operator: calling object is multiplied by fraction on the right
						   of the operator.

					Calling object must be of type %MFraction. The calling object is the fraction that is on the left
					of the operator.

					@param right %MFraction object on the right of operator that is to be multiplied with. If values of
						   integer, double, strings or %Fraction are provided, they will be automatically converted to
						   %MFraction objects prior to multiplication/assignment.
					@return Returns a fraction object that is equal to the value of the updated calling object.
				*/
				MFraction operator *= (const MFraction &right);

				/*! @brief Divition/assignment operator: calling object is divided by fraction on the right
						   of the operator.

					Calling object must be of type %MFraction. The calling object is the fraction that is on the left
					of the operator.

					@param right %MFraction object on the right of operator that is to be divided by. If values of
						   integer, double, strings or %Fraction are provided, they will be automatically converted to
						   %Fraction objects prior to division/assignment.
					@return Returns a fraction object that is equal to the value of the updated calling object.
				*/
				MFraction operator /= (const MFraction &right);

				//@}

				/* FRIENDS: OVERLOADED OPERATORS */

				/** @name Comparison Operations
					These are declared as friends so that it is possible to 25 < x or 5==x where x is a %MFraction
					object. Parameters are automatically converted to %MFraction objects when values of integer,
					double, string or Fraction objects are provided. However, note that at least one of the parameters
					must be an object of type %MFraction. */
				//@{

				//! Equality operator: checks if two fractions are same
				/*! @param left %MFraction object to the left of the operator.
					@param right %MFraction object to the right of the operator.
					@return Returns true if the two fraction objects are equal. False otherwise.
				*/
				friend bool operator == (const MFraction &left, const MFraction &right);
				
				//! Inequality operator: checks if two fractions arent equal
				/*! @param left %MFraction object to the left of the operator.
					@param right %MFraction object to the right of the operator.
					@return Returns true if the two fraction objects aren't equal. False otherwise.
				*/
				friend bool operator != (const MFraction &left, const MFraction &right);
			
				//! Less than operator: checks if fraction on the left of operator is less than fraction on the right.
				/*! @param left %MFraction object to the left of the operator.
					@param right %MFraction object to the right of the operator.
					@return Returns true if left fraction is less than right fraction. False otherwise.
				*/
				friend bool operator < (const MFraction &left, const MFraction &right);
				
				/*! @brief Less than or equal to operator: checks if fraction on the left of operator is less than
						   or equal to fraction on the right.
				
					@param left %MFraction object to the left of the operator.
					@param right %MFraction object to the right of the operator.
					@return Returns true if left fraction is less than or equal to right fraction. False otherwise.
				*/
				friend bool operator <= (const MFraction &left, const MFraction &right);
				
				/*! @brief Greater than operator: checks if fraction on the left of operator is greater than
						   fraction on the right.
				
					@param left %MFraction object to the left of the operator.
					@param right %MFraction object to the right of the operator.
					@return Returns true if left fraction is greater than right fraction. False otherwise.
				*/
				friend bool operator > (const MFraction &left, const MFraction &right);

				/*! @brief Greater than or equal to operator: checks if fraction on the left of operator is greater than
						   or equal to fraction on the right.
				
					@param left %MFraction object to the left of the operator.
					@param right %MFraction object to the right of the operator.
					@return Returns true if left fraction is greater than or equal to right fraction. False otherwise.
				*/
				friend bool operator >= (const MFraction &left, const MFraction &right);

				//@}

				/** @name Arithmitic Operations
					These are declared as friends so that it is possible to 25 + x or 5/x where x is a %MFraction
					object. Parameters are automatically converted to %MFraction objects when values of integer,
					double, string or %Fraction objects are provided. However, note that at least one of the parameters
					must be an object of type %MFraction.
				*/
				//@{

				/*! @brief Addition operator: adds the two fractions on either side of the operator.
					@param left %MFraction on the left of the operator
					@param right %MFraction on the right of the operator
					@return Returns the reduced sum of the two fractions.
				*/
				friend MFraction operator + (const MFraction &left, const MFraction &right);
				
				/*! @brief Subtraction operator: subtracts the two fractions on either side of the operator.
					@param left %MFraction on the left of the operator
					@param right %MFraction on the right of the operator
					@return Returns the reduced difference of the two fractions.
				*/
				friend MFraction operator - (const MFraction &left, const MFraction &right);
				
				/*! @brief Multiplication operator: multiplies the two fractions on either side of the operator.
					@param left %MFraction on the left of the operator
					@param right %MFraction on the right of the operator
					@return Returns the reduced product of the two fractions.
				*/
				friend MFraction operator * (const MFraction &left, const MFraction &right);
				
				/*! @brief Division operator: left fraction is divided by the fraction on the right of the operator.
					@param left %MFraction on the left of the operator
					@param right %MFraction on the right of the operator
					@return Returns the reduced quotient of the two fractions.
				*/
				friend MFraction operator / (const MFraction &left, const MFraction &right);

				//@}

				/** @name Stream input/output operations */
				//@{		

				/*! @brief Stream output: output fraction in the format specified by MFraction::FORMAT
					@param out output stream (operation assumes stream is already setup (ie. connected to file, screen etc.)
					@param fraction %MFraction object to output.
					@return Returns reference to output stream so that for instance, cout << x << y is possible.
				*/
				friend ostream &operator << (ostream &out, const MFraction &fraction);
				
				/*! @brief Stream input: assign input to a fraction object
					@param in input stream (operation assumes stream is already setup (ie. connected to file, screen etc.)
					@param fraction The input from stream is assigned to this fraction object.
					@return Returns reference to input stream so that for instance, cin >> x >> y is possible.
				*/
				friend istream &operator >> (istream &in, MFraction &fraction);

				//@}
		};
	#endif /* #ifndef FRACTION_ONLY */
#endif /* #ifndef MFRACTION_H */